@@ -2,6 +2,7 @@
 # SPDX-License-Identifier: BSD-2-Clause
 
 # For CRAN
+oldThreads <- getDelapThreads()
 setDelapThreads(2L)
 
 tol <- 5e-4
@@ -56,7 +57,7 @@ expect_identical(length(rdelap(8, c(4, 2), c(1, 2, 3, 4), 2)), 8L)
 expect_identical(length(rdelap(0, c(4, 2), c(1, 2, 3, 4), 2)), 0L)
 expect_error(rdelap(-1, c(4, 2), c(1, 2, 3, 4), 2), nlErr)
 
-# Approximate throws error when nonpositive is passed
+# Approximate throws warning when nonpositive is passed
 expect_warning(rdelap(8, 0, 2, 3, exact = FALSE), nanWarn)
 expect_warning(rdelap(8, 1, -0.8, 3, exact = FALSE), nanWarn)
 expect_warning(rdelap(8, 1, 2, 0, exact = FALSE), nanWarn)
@@ -67,3 +68,6 @@ INTG <- rdelap(3, 1L, 2L, 3L)
 set.seed(17L)
 DOUBL <- rdelap(3, 1, 2, 3)
 expect_equal(INTG, DOUBL, tolerance = 1e-12)
+
+# Restore original thread count
+setDelapThreads(oldThreads)
