@@ -39,6 +39,22 @@
 !                       Added OpenMP thread control functionality.
 !          Version 4.1: 2024-04-04
 !                       Use "source" when allocating arrays in qdelap_f.
+!          Version 5.0: 2024-06-17
+!                       OpenMP is still significantly faster than extending
+!                       parameter vectors and applying the elemental singleton
+!                       function. Use imk helper function to calculate position
+!                       for vector recycling. Turn floating point error cleaner
+!                       into a function. Move lt and lg for p/delap into
+!                       existing loops. While elemental functions can be even
+!                       faster than OpenMP loops, they are still another loop.
+!                       Saving the overhead by calling the conditionals inside
+!                       the necessary loops is still faster. This is not useful
+!                       for qdelap as we need to know every "real" percentile
+!                       first. So run those before anything else. Other minor
+!                       tweaks to code for efficiency. SIMD instructions do not
+!                       save that much time and prevent compilation with current
+!                       versions of flang (through 19) as it does not have full
+!                       OpenMP 4.5 implementation yet.
 !
 ! LICENSE:
 !   Copyright (c) 2016, Avraham Adler
@@ -86,18 +102,20 @@ contains
 !              explicit summation. Follows R convention that real observations
 !              are errors and have 0 probability, so calls floor to build to
 !              the last integer. Implements hard floor of 0 and hard ceiling of
-!              1 to prevent spurious floating point errors.
+!              1 to prevent spurious floating point errors. The single vector
+!              trick from pdelap actually slows ddelap down in almost all case
+!              unless the passed vectors are very close to one another and small
+!              in magnitude, so it is not worth programming for now.
 !-------------------------------------------------------------------------------
 
-    elemental function ddelap_f_s(x, alpha, beta, lambda) result(pmf)
+    pure elemental function ddelap_f_s(x, alpha, beta, lambda) result(pmf)
 
     real(kind = c_double), intent(in)   :: x, alpha, beta, lambda
     real(kind = c_double)               :: pmf, ii, kk
     integer(INT64)                      :: i, k                   
 
         if (alpha <= ZERO .or. beta <= ZERO .or. lambda <= ZERO .or. x < ZERO &
-            .or. ieee_is_nan(alpha) .or. ieee_is_nan(beta) .or. &
-            ieee_is_nan(lambda) .or. ieee_is_nan(x)) then
+            .or. ieee_is_nan(alpha + beta + lambda + x)) then
             pmf = ieee_value(x, ieee_quiet_nan)
         else
             pmf = ZERO
@@ -111,7 +129,7 @@ contains
                     - log_gamma(ii + ONE) - (alpha + ii) * log1p(beta) &
                     - log_gamma(kk - ii + ONE))
                 end do
-            pmf = max(min(pmf, ONE), ZERO)        ! Clear floating point errors
+            pmf = cFPe(pmf)                       ! Clear floating point errors
             end if
         end if
 
@@ -132,31 +150,25 @@ contains
     subroutine ddelap_f(x, nx, a, na, b, nb, l, nl, lg, threads, pmfv) &
                bind(C, name="ddelap_f_")
                         
-    integer(kind = c_int), intent(in), value         :: nx, na, nb, nl
-    real(kind = c_double), intent(in), dimension(nx) :: x
-    real(kind = c_double), intent(out), dimension(nx):: pmfv
-    real(kind = c_double), intent(in)                :: a(na), b(nb), l(nl)
-    integer(kind = c_int), intent(in)                :: lg, threads
-    integer                                          :: i
+    integer(kind = c_int), intent(in), value     :: nx, na, nb, nl
+    real(kind = c_double), intent(in)            :: x(nx), a(na), b(nb), l(nl)
+    integer(kind = c_int), intent(in)            :: lg, threads
+    real(kind = c_double), intent(out)           :: pmfv(nx)
+    integer                                      :: i
     
         !$omp parallel do num_threads(threads) default(shared) private(i) &
         !$omp schedule(static)
         do i = 1, nx
-            pmfv(i) = ddelap_f_s(x(i), a(mod(i - 1, na) + 1), &
-            b(mod(i - 1, nb) + 1), l(mod(i - 1, nl) + 1))
+            pmfv(i) = ddelap_f_s(x(i), a(imk(i, na)), b(imk(i, nb)), &
+            l(imk(i, nl)))
+            if (lg == 1) pmfv(i) = log(pmfv(i))
         end do
         !$omp end parallel do
         
-        if (lg == 1) then
-            pmfv = log(pmfv)
-        end if
-        
-        if (any(ieee_is_nan(pmfv))) then
-            call rwarn("NaNs produced")
-        end if
-        
-    end subroutine ddelap_f
+        if (any(ieee_is_nan(pmfv))) call rwarn("NaNs produced")
 
+    end subroutine ddelap_f
+    
 !-------------------------------------------------------------------------------
 ! FUNCTION: pdelap_f_s
 !
@@ -168,15 +180,14 @@ contains
 !              ceiling of 1 to prevent spurious floating point errors.
 !-------------------------------------------------------------------------------
 
-    elemental function pdelap_f_s(q, alpha, beta, lambda) result(cdf)
+    pure elemental function pdelap_f_s(q, alpha, beta, lambda) result(cdf)
 
-    real(kind = c_double)               :: cdf
     real(kind = c_double), intent(in)   :: q, alpha, beta, lambda
+    real(kind = c_double)               :: cdf
     integer(INT64)                      :: i, k
 
         if (alpha <= ZERO .or. beta <= ZERO .or. lambda <= ZERO .or. q < ZERO &
-            .or. ieee_is_nan(alpha) .or. ieee_is_nan(beta) .or. &
-            ieee_is_nan(lambda) .or. ieee_is_nan(q)) then
+            .or. ieee_is_nan(alpha + beta + lambda + q)) then
             cdf = ieee_value(q, ieee_quiet_nan)
         else if (.not. ieee_is_finite(q)) then
             cdf = ONE
@@ -186,7 +197,7 @@ contains
             do i = 1_INT64, k
                 cdf = cdf + ddelap_f_s(real(i, c_double), alpha, beta, lambda)
             end do
-        cdf = max(min(cdf, ONE), ZERO)        ! Clear floating point errors
+            cdf = cFPe(cdf)                       ! Clear floating point errors
         end if
 
     end function pdelap_f_s
@@ -200,7 +211,7 @@ contains
 !              up to that point. Building the vector has each succesive value
 !              piggyback off of the prior instead of calling p_delap_f_s each
 !              time which increases the speed dramatically. Once created,
-!              remaining values are simple lookups off of the singlevec vector.
+!              remaining values are simple lookups off of the svec vector.
 !              Otherwise, each entry will need to build its own pmf value by
 !              calling p_delap_f_s on each entry. Implements hard floor of 0 and
 !              hard ceiling of 1 to prevent spurious floating point errors.
@@ -209,16 +220,15 @@ contains
     subroutine pdelap_f(q, nq, a, na, b, nb, l, nl, lt, lg, threads, pmfv) &
                bind(C, name="pdelap_f_")
                         
-    integer(kind = c_int), intent(in), value         :: nq, na, nb, nl
-    real(kind = c_double), intent(in), dimension(nq) :: q
-    real(kind = c_double), intent(out), dimension(nq):: pmfv
-    real(kind = c_double), intent(in)                :: a(na), b(nb), l(nl)
-    integer(kind = c_int), intent(in)                :: lg, lt, threads
-    real(kind = c_double), allocatable, dimension(:) :: singlevec
-    integer                                          :: i, k
+    integer(kind = c_int), intent(in), value    :: nq, na, nb, nl
+    real(kind = c_double), intent(in)           :: q(nq), a(na), b(nb), l(nl)
+    integer(kind = c_int), intent(in)           :: lg, lt, threads
+    real(kind = c_double), intent(out)          :: pmfv(nq)
+    real(kind = c_double), allocatable          :: svec(:)
+    integer                                     :: i, k
 
 ! If there are any complications at all, don't use the fast version. pdelap_f_s
-! and ddelap_f_s are more robust to improper entries
+! and ddelap_f_s are more robust to improper entries.
 
         if (na > 1 .or. nb > 1 .or. nl > 1 .or. minval(q) < ZERO .or. &
             maxval(q) > REAL(MAXVECSIZE, c_double) &
@@ -226,43 +236,32 @@ contains
             !$omp parallel do num_threads(threads) default(shared) private(i) &
             !$omp schedule(static)
                 do i = 1, nq
-                    pmfv(i) = pdelap_f_s(q(i), a(mod(i - 1, na) + 1), &
-                    b(mod(i - 1, nb) + 1), l(mod(i - 1, nl) + 1))
+                    pmfv(i) = pdelap_f_s(q(i), a(imk(i, na)), b(imk(i, nb)), &
+                    l(imk(i, nl)))
+                    if (lt == 0) pmfv(i) = HALF - pmfv(i) + HALF ! See See dpq.h
+                    if (lg == 1) pmfv(i) = log(pmfv(i))
                 end do
             !$omp end parallel do
+        else if (a(1) <= ZERO .or. b(1) <= ZERO .or. l(1) <= ZERO .or. &
+                 ieee_is_nan(a(1) + b(1) + l(1))) then
+            pmfv = ieee_value(q, ieee_quiet_nan)
         else
-            if (a(1) <= ZERO .or. b(1) <= ZERO .or. l(1) <= ZERO .or. &
-                ieee_is_nan(a(1)) .or. ieee_is_nan(b(1)) .or. &
-                ieee_is_nan(l(1))) then
-                pmfv = ieee_value(q, ieee_quiet_nan)
-            else
-                k = floor(maxval(q))
-                allocate (singlevec(k + 1))
-                singlevec(1) = exp(-l(1)) / ((b(1) + ONE) ** a(1))
-                do i = 2, k + 1
-                    singlevec(i) = singlevec(i - 1) &
-                    + ddelap_f_s(real(i - 1, c_double), a(1), b(1), l(1))
-                end do
-                do i = 1, nq
-                    k = floor(q(i))
-                    pmfv(i) = singlevec(k + 1)
-                end do
-                deallocate(singlevec)
-                pmfv = max(min(pmfv, ONE), ZERO) ! Clear floating point errors
-            end if
+            k = floor(maxval(q))
+            allocate (svec(k + 1))
+            svec(1) = cFPe(exp(-l(1)) / ((b(1) + ONE) ** a(1)))
+            do i = 2, k + 1
+                svec(i) = cFPe(svec(i - 1) + &
+                ddelap_f_s(real(i - 1, c_double), a(1), b(1), l(1)))
+            end do
+            do i = 1, nq
+                pmfv(i) = svec(floor(q(i)) + 1)
+                if (lt == 0) pmfv(i) = HALF - pmfv(i) + HALF  ! See See dpq.h
+                if (lg == 1) pmfv(i) = log(pmfv(i))
+            end do
+            deallocate(svec)
         end if
         
-        if (lt == 0) then
-            pmfv = ONE - pmfv
-        end if
-        
-        if (lg == 1) then
-            pmfv = log(pmfv)
-        end if
-        
-        if (any(ieee_is_nan(pmfv))) then
-            call rwarn("NaNs produced")
-        end if
+        if (any(ieee_is_nan(pmfv))) call rwarn("NaNs produced")
         
     end subroutine pdelap_f
 
@@ -274,14 +273,13 @@ contains
 !              summation. Returns NaN and Inf where appropriate.
 !-------------------------------------------------------------------------------
 
-    elemental function qdelap_f_s(p, alpha, beta, lambda) result(value)
+    pure elemental function qdelap_f_s(p, alpha, beta, lambda) result(value)
 
     real(kind = c_double), intent(in)   :: p, alpha, beta, lambda
     real(kind = c_double)               :: testcdf, value
 
         if (alpha <= ZERO .or. beta <= ZERO .or. lambda <= ZERO .or. p < ZERO &
-          .or. ieee_is_nan(alpha) .or. ieee_is_nan(beta) .or. &
-          ieee_is_nan(lambda) .or. ieee_is_nan(p)) then
+          .or. ieee_is_nan(alpha + beta + lambda + p)) then
             value = ieee_value(p, ieee_quiet_nan)
         else if (p >= ONE) then
             value = ieee_value(p, ieee_positive_inf)
@@ -305,7 +303,7 @@ contains
 !              that point. Building the vector has each succesive value
 !              piggyback off of the prior instead of calling p_delap_f_s each
 !              time which increases the speed dramatically. Once created,
-!              remaining values are lookups off of the singlevec vector.
+!              remaining values are lookups off of the svec vector.
 !              Otherwise, each entry will need to build its own pmf value by
 !              calling q_delap_f_s on each entry.
 !-------------------------------------------------------------------------------
@@ -313,22 +311,18 @@ contains
     subroutine qdelap_f(p, np, a, na, b, nb, l, nl, lt, lg, threads, obsv) &
                bind(C, name="qdelap_f_")
 
-    integer(kind = c_int), intent(in), value           :: np, na, nb, nl
-    real(kind = c_double), intent(inout), dimension(np):: p
-    real(kind = c_double), intent(out), dimension(np)  :: obsv
-    real(kind = c_double), intent(in)                  :: a(na), b(nb), l(nl)
-    integer(kind = c_int), intent(in)                  :: lg, lt, threads
-    real(kind = c_double), allocatable, dimension(:)   :: svec, tvec
-    real(kind = c_double)                              :: x
-    integer                                            :: i
+    integer(kind = c_int), intent(in), value        :: np, na, nb, nl
+    real(kind = c_double), intent(in)               :: a(na), b(nb), l(nl)
+    integer(kind = c_int), intent(in)               :: lg, lt, threads
+    real(kind = c_double), intent(inout)            :: p(np) 
+    real(kind = c_double), intent(out)              :: obsv(np)
+    real(kind = c_double), allocatable              :: svec(:), tvec(:)
+    real(kind = c_double)                           :: x
+    integer                                         :: i
 
-        if (lg == 1) then
-            p = exp(p)
-        end if
+        if (lg == 1) p = exp(p)
 
-        if (lt == 0) then
-            p = ONE - p
-        end if
+        if (lt == 0) p = HALF - p + HALF       ! See See dpq.h in R source code
 
         if(na == 1 .and. nb == na .and. nl == nb) then
             if (a(1) <= ZERO .or. b(1) <= ZERO .or. l(1) <= ZERO) then
@@ -338,15 +332,13 @@ contains
                 allocate(svec(1), source = exp(-l(1)) / ((b(1) + ONE) ** a(1)))
                 i = 1
                 do
-                    if (svec(i) >= x) then
-                        exit
-                    end if
+                    if (svec(i) >= x) exit
                     i = i + 1
                     allocate(tvec(1:i), source = ZERO)
                     tvec(1:i-1) = svec
                     call move_alloc(tvec, svec)
-                    svec(i) = svec(i - 1) + ddelap_f_s(real(i - 1, c_double), &
-                                                       a(1), b(1), l(1))
+                    svec(i) = svec(i - 1) + &
+                        ddelap_f_s(real(i - 1, c_double), a(1), b(1), l(1))
                 end do
                 do i = 1, np
                     if (p(i) < ZERO .or. ieee_is_nan(p(i))) then
@@ -364,9 +356,8 @@ contains
             !$omp parallel do num_threads(threads) default(shared) private(i) &
             !$omp schedule(static)
             do i = 1, np
-                obsv(i) = qdelap_f_s(p(i), a(mod(i - 1, na) + 1), &
-                                     b(mod(i - 1, nb) + 1), &
-                                     l(mod(i - 1, nl) + 1))
+                obsv(i) = qdelap_f_s(p(i), a(imk(i, na)), b(imk(i, nb)), &
+                          l(imk(i, nl)))
             end do
             !$omp end parallel do
         end if
@@ -397,15 +388,13 @@ contains
     external unifrnd
 
     integer(kind = c_int), intent(in), value           :: n, na, nb, nl
-    real(kind = c_double), intent(out), dimension(n)   :: vars
     real(kind = c_double), intent(in)                  :: a(na), b(nb), l(nl)
-    real(kind = c_double), dimension(n)                :: p
-    integer(kind = c_int)                              :: lg, lt, threads
+    real(kind = c_double), intent(out)                 :: vars(n)
+    real(kind = c_double)                              :: p(n)
+    integer(kind = c_int)                              :: threads
 
         call unifrnd(n, p)
-        lt = 1_c_int
-        lg = 0_c_int
-        call qdelap_f(p, n, a, na, b, nb, l, nl, lt, lg, threads, vars)
+        call qdelap_f(p, n, a, na, b, nb, l, nl, 1, 0, threads, vars)
 
     end subroutine rdelap_f
 
@@ -418,29 +407,28 @@ contains
 !              https://www.johndcook.com/blog/skewness_kurtosis/
 !-------------------------------------------------------------------------------
 
-    subroutine momdelap_f(obs, n, tp, params) bind(C, name="momdelap_f_")
+    pure subroutine momdelap_f(obs, n, tp, params) bind(C, name="momdelap_f_")
 
-    integer(kind = c_int), intent(in), value           :: n
-    integer(kind = c_int), intent(in)                  :: tp
-    real(kind = c_double), intent(in), dimension(n)    :: obs
-    real(kind = c_double), intent(out), dimension(3)   :: params
-    real(kind = c_double)                              :: nm1, P, Mu_D, M2, M3
-    real(kind = c_double)                              :: T1, delta, delta_i, nr
-    real(kind = c_double)                              :: Var_D, Skew_D, VmM_D
-    real(kind = c_double)                              :: ii
-    integer                                            :: i
+    integer(kind = c_int), intent(in), value :: n
+    integer(kind = c_int), intent(in)        :: tp
+    real(kind = c_double), intent(in)        :: obs(n)
+    real(kind = c_double), intent(out)       :: params(3)
+    real(kind = c_double)                    :: nnm1, P, Mu_D, M2, M3, T1
+    real(kind = c_double)                    :: delta, delta_i, nn, Var_D
+    real(kind = c_double)                    :: Skew_D, VmM_D, ii
+    integer                                  :: i
 
-        nr = real(n, c_double)
-        nm1 = nr - ONE
+        nn = real(n, c_double)
+        nnm1 = nn - ONE
         select case (tp)
             case (1)
                 P = ONE
             case (2)
-                P = sqrt(nr * nm1) / (nr - TWO)
+                P = sqrt(nn * nnm1) / (nn - TWO)
             case (3)
-                P = (nm1 / nr) ** THREEHALFS
+                P = (nnm1 / nn) ** THREEHALFS
             case default
-                P = sqrt(nr * nm1) / (nr - TWO)
+                P = sqrt(nn * nnm1) / (nn - TWO)
         end select
         Mu_D = ZERO
         M2 = ZERO
@@ -454,8 +442,8 @@ contains
             M3 = M3 + (T1 * delta_i * (ii - TWO) - THREE * delta_i * M2)
             M2 = M2 + T1
         end do
-        Var_D = M2 / nm1
-        Skew_D = P * sqrt(nr) * M3 / (M2 ** THREEHALFS)
+        Var_D = M2 / nnm1
+        Skew_D = P * sqrt(nn) * M3 / (M2 ** THREEHALFS)
         VmM_D = Var_D - Mu_D
         params(2) = HALF * (Skew_D * (Var_D ** THREEHALFS) - Mu_D - THREE &
                             * VmM_D) / VmM_D
